Class: Bitcoin::PartialTree::Node
- Inherits:
-
Object
- Object
- Bitcoin::PartialTree::Node
- Defined in:
- lib/bitcoin/partial_tree.rb
Overview
node of merkle tree
Instance Attribute Summary collapse
-
#flag ⇒ Object
Returns the value of attribute flag.
-
#left ⇒ Object
Returns the value of attribute left.
-
#parent ⇒ Object
Returns the value of attribute parent.
-
#right ⇒ Object
Returns the value of attribute right.
-
#value ⇒ Object
Returns the value of attribute value.
Instance Method Summary collapse
-
#depth ⇒ Object
calculate the depth of this node in the tree.
-
#find_node(target) ⇒ Object
Node which has same value as target.
- #index ⇒ Object
-
#initialize(value = nil) ⇒ Node
constructor
A new instance of Node.
- #leaf? ⇒ Boolean
- #next_partial ⇒ Object
- #partial? ⇒ Boolean
- #root? ⇒ Boolean
Constructor Details
#initialize(value = nil) ⇒ Node
Returns a new instance of Node.
90 91 92 |
# File 'lib/bitcoin/partial_tree.rb', line 90 def initialize(value = nil) @value = value end |
Instance Attribute Details
#flag ⇒ Object
Returns the value of attribute flag.
84 85 86 |
# File 'lib/bitcoin/partial_tree.rb', line 84 def flag @flag end |
#left ⇒ Object
Returns the value of attribute left.
87 88 89 |
# File 'lib/bitcoin/partial_tree.rb', line 87 def left @left end |
#parent ⇒ Object
Returns the value of attribute parent.
86 87 88 |
# File 'lib/bitcoin/partial_tree.rb', line 86 def parent @parent end |
#right ⇒ Object
Returns the value of attribute right.
88 89 90 |
# File 'lib/bitcoin/partial_tree.rb', line 88 def right @right end |
#value ⇒ Object
Returns the value of attribute value.
85 86 87 |
# File 'lib/bitcoin/partial_tree.rb', line 85 def value @value end |
Instance Method Details
#depth ⇒ Object
calculate the depth of this node in the tree.
131 132 133 134 135 136 137 138 139 |
# File 'lib/bitcoin/partial_tree.rb', line 131 def depth d = 0 current_node = self until current_node.root? do current_node = current_node.parent d += 1 end d end |
#find_node(target) ⇒ Object
Returns node which has same value as target. nil if this node and any children don't have same value.
143 144 145 146 147 |
# File 'lib/bitcoin/partial_tree.rb', line 143 def find_node(target) return self if value == target return nil if flag && flag.zero? return left&.find_node(target) || right&.find_node(target) end |
#index ⇒ Object
149 150 151 152 153 154 155 156 157 158 159 |
# File 'lib/bitcoin/partial_tree.rb', line 149 def index i = 0 d = 1 current_node = self until current_node.root? do i += d if current_node.parent.right == current_node current_node = current_node.parent d *= 2 end i end |
#leaf? ⇒ Boolean
114 115 116 |
# File 'lib/bitcoin/partial_tree.rb', line 114 def leaf? right.nil? && left.nil? end |
#next_partial ⇒ Object
122 123 124 125 126 127 128 |
# File 'lib/bitcoin/partial_tree.rb', line 122 def next_partial return nil if root? && (flag.zero? || (left.nil? && right.nil?) || (left.partial? && right.partial?)) return parent.next_partial if flag.zero? || leaf? return left unless left.partial? self.right = left.dup unless right right.partial? ? parent.next_partial : right end |
#partial? ⇒ Boolean
118 119 120 |
# File 'lib/bitcoin/partial_tree.rb', line 118 def partial? !flag.nil? end |
#root? ⇒ Boolean
110 111 112 |
# File 'lib/bitcoin/partial_tree.rb', line 110 def root? parent.nil? end |